#!/usr/bin/env python3
#
#  decoder.py
"""
TOML decoders.

.. versionadded:: 0.2.0
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Based on https://github.com/uiri/toml
#  MIT Licensed
#  Copyright 2013-2019 William Pearson
#  Copyright 2015-2016 Julien Enselme
#  Copyright 2016 Google Inc.
#  Copyright 2017 Samuel Vasko
#  Copyright 2017 Nate Prewitt
#  Copyright 2017 Jack Evans
#  Copyright 2019 Filippo Broggini
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import sys
from typing import Any, Callable, Dict, Tuple

if sys.version_info >= (3, 11):  # pragma: no cover (<py311)
	# stdlib
	import tomllib as tomli
else:  # pragma: no cover (py311+)
	# 3rd party
	import tomli

__all__ = ["InlineTableDict", "TomlDecoder", "TomlPureDecoder"]


class InlineTableDict(dict):
	"""
	Subclass of dict for inline tables.

	.. versionadded:: 2.0.0
	"""


class TomlDecoder:
	"""
	TOML decoder which uses a dict-subclass for inline tables.

	.. versionadded:: 2.0.0
	"""

	def loads(self, s: str) -> Dict[str, Any]:
		"""
		Parse the given string as TOML.

		:param s:

		:returns: A mapping containing the ``TOML`` data.

		.. latex:clearpage::
		"""

		try:
			pit = tomli._parser.parse_inline_table

			def _parse_inline_table(
					src: str,
					pos: int,
					parse_float: Callable[[str], Any],
					*args,
					**kwargs,
					) -> Tuple[int, Dict]:
				pos, table = pit(src, pos, parse_float, *args, **kwargs)
				return pos, InlineTableDict(table)

			tomli._parser.parse_inline_table = _parse_inline_table
			return tomli.loads(s)
		finally:
			tomli._parser.parse_inline_table = pit


class TomlPureDecoder(TomlDecoder):
	"""
	TOML decoder which uses pure-Python dictionaries for inline tables.
	"""

	def loads(self, s: str) -> Dict[str, Any]:
		"""
		Parse the given string as TOML.

		:param s:

		:returns: A mapping containing the ``TOML`` data.

		.. versionadded:: 2.0.0
		"""

		return tomli.loads(s)



#!/usr/bin/env python3
#
#  parser.py
"""
Abstract base class for TOML configuration parsers.

.. versionadded:: 0.2.0
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from abc import ABC, abstractmethod
from typing import Any, Callable, ClassVar, Dict, Iterable, List, Optional, Tuple, Type, TypeVar, Union

# this package
import dom_toml

__all__ = ["AbstractConfigParser", "BadConfigError", "construct_path", "TOML_TYPES"]

TOML_TYPES = Any


class BadConfigError(ValueError):
	"""
	Indicates an error in the TOML configuration.

	:param documentation: A link to the documentation that explains the problematic option.
		This is not used by the class itself, except setting it as the ``documentation`` attribute,
		The intention is for code catching this exception to display this URL to the user.

	.. versionchanged:: 0.6.0  Added the ``documentation`` keyword argument and attribute.
	"""

	#: A link to the documentation that explains the problematic option.
	documentation: Optional[str]

	def __init__(self, *args, documentation: Optional[str] = None) -> None:
		super().__init__(*args)
		self.documentation = documentation


def construct_path(path: Iterable[str]) -> str:
	"""
	Construct a dotted path to a key.

	:param path: The path elements.
	"""

	return '.'.join([dom_toml.dumps({elem: 0})[:-5] for elem in path])


_C = TypeVar("_C", bound=Callable)


class AbstractConfigParser(ABC):
	"""
	Abstract base class for TOML configuration parsers.

	.. autoclasssumm:: AbstractConfigParser
	.. latex:clearpage::
	"""

	defaults: ClassVar[Dict[str, Any]]
	"""
	A mapping of key names to default values.

	.. versionadded:: 0.3.0
	"""

	factories: ClassVar[Dict[str, Callable[..., Any]]]
	"""
	A mapping of key names to default value factories.

	.. versionadded:: 0.3.0

	.. note:: If both a default and a factory are defined for a key the factory takes precedence.

	.. note::

		``defaults`` and ``factories`` are reset for each subclass.
		To disable this behaviour set the ``inherit_defaults`` keyword argument on the class:

		.. code-block:: python

			class MyParser(AbstractConfigParser, inherit_default=True):
				pass
	"""

	def __init_subclass__(cls, **kwargs) -> None:
		if not kwargs.get("inherit_defaults", False):
			if "defaults" not in cls.__dict__:
				cls.defaults = {}

			if "factories" not in cls.__dict__:
				cls.factories = {}

	@staticmethod
	def assert_type(
			obj: Any,
			expected_type: Union[Type, Tuple[Type, ...]],
			path: Iterable[str],
			what: str = "type",
			) -> None:
		"""
		Assert that ``obj`` is of type ``expected_type``, otherwise raise an error with a helpful message.

		:param obj: The object to check the type of.
		:param expected_type: The expected type.
		:param path: The elements of the path to ``obj`` in the TOML mapping.
		:param what: What ``obj`` is, e.g. ``'type'``, ``'value type'``.

		.. seealso:: :meth:`~.assert_value_type` and :meth:`~.assert_indexed_type`
		"""

		if not isinstance(obj, expected_type):
			name = construct_path(path)
			raise TypeError(f"Invalid {what} for {name!r}: expected {expected_type!r}, got {type(obj)!r}")

	@staticmethod
	def assert_indexed_type(
			obj: Any,
			expected_type: Union[Type, Tuple[Type, ...]],
			path: Iterable[str],
			idx: int = 0,
			) -> None:
		"""
		Assert that ``obj`` is of type ``expected_type``, otherwise raise an error with a helpful message.

		:param obj: The object to check the type of.
		:param expected_type: The expected type.
		:param path: The elements of the path to ``obj`` in the TOML mapping.
		:param idx: The index of ``obj`` in the array.

		.. seealso:: :meth:`~.assert_type`, and :meth:`~.assert_value_type`
		"""

		if not isinstance(obj, expected_type):
			name = construct_path(path) + f"[{idx}]"
			raise TypeError(f"Invalid type for {name!r}: expected {expected_type!r}, got {type(obj)!r}")

	def assert_value_type(
			self,
			obj: Any,
			expected_type: Union[Type, Tuple[Type, ...]],
			path: Iterable[str],
			) -> None:
		"""
		Assert that the value ``obj`` is of type ``expected_type``, otherwise raise an error with a helpful message.

		:param obj: The object to check the type of.
		:param expected_type: The expected type.
		:param path: The elements of the path to ``obj`` in the TOML mapping.

		.. seealso:: :meth:`~.assert_type` and :meth:`~.assert_indexed_type`
		"""

		self.assert_type(obj, expected_type, path, "value type")

	@property
	@abstractmethod
	def keys(self) -> List[str]:  # pragma: no cover
		"""
		The keys to parse from the TOML file.
		"""

		raise NotImplementedError

	def parse(
			self,
			config: Dict[str, TOML_TYPES],
			set_defaults: bool = False,
			) -> Dict[str, TOML_TYPES]:
		r"""
		Parse the TOML configuration.

		This function iterates over the list of keys given in :attr:`~.keys`.
		For each key, it searches for a method on the class called :file:`parse_{<key>}`.

		* If the method exists, that method is called, passing the value as the only argument.
		  The value returned from that method is included in the parsed configuration.
		  The signature of those methods is:

		  .. parsed-literal::

			def visit_<key>(
				self,
				config: :class:`typing.Dict`\[:class:`str`\, :py:obj:`typing.Any`\],
				) -> :py:obj:`typing.Any`\:

		* If the method doesn't exist, the value is included in the parsed configuration unchanged.

		* Missing keys are ignored. Override this function in a subclass if you need that behaviour.

		Once all keys have been parsed the configuration is returned.

		:param config:
		:param set_defaults: If :py:obj:`True`, the values in :attr:`.AbstractConfigParser.defaults`
			and :attr:`.AbstractConfigParser.factories` will be set as defaults for the returned mapping.

		.. versionchanged:: 0.3.0

			Added the ``set_defaults`` keyword argument.
		"""

		parsed_config = {}

		for key in self.keys:
			if key not in config:
				# Ignore absent values
				pass

			elif hasattr(self, f"parse_{key.replace('-', '_')}"):
				parsed_config[key] = getattr(self, f"parse_{key.replace('-', '_')}")(config)

			elif key in config:
				parsed_config[key] = config[key]

		if set_defaults:
			for key, value in self.defaults.items():
				parsed_config.setdefault(key, value)

			for key, factory in self.factories.items():
				value = factory()
				parsed_config.setdefault(key, value)

		return parsed_config



#!/usr/bin/env python3
#
#  __init__.py
"""
Dom's tools for Tom's Obvious, Minimal Language.
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Based on https://github.com/uiri/toml
#  MIT Licensed
#  Copyright 2013-2019 William Pearson
#  Copyright 2015-2016 Julien Enselme
#  Copyright 2016 Google Inc.
#  Copyright 2017 Samuel Vasko
#  Copyright 2017 Nate Prewitt
#  Copyright 2017 Jack Evans
#  Copyright 2019 Filippo Broggini
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Any, Dict, Mapping, Type, Union

# 3rd party
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike

# this package
from dom_toml.decoder import TomlDecoder
from dom_toml.encoder import TomlEncoder

__author__: str = "Dominic Davis-Foster"
__copyright__: str = "2021 Dominic Davis-Foster"
__license__: str = "MIT License"
__version__: str = "2.1.0"
__email__: str = "dominic@davis-foster.co.uk"

__all__ = ["dumps", "loads", "dump", "load"]


def dumps(
		data: Mapping[str, Any],
		encoder: Union[Type[TomlEncoder], TomlEncoder] = TomlEncoder,
		) -> str:
	r"""
	Convert ``data`` to a TOML string.

	:param data:
	:param encoder: The :class:`~.TomlEncoder` to use for constructing the output string.

	:returns: A string containing the ``TOML`` corresponding to ``data``.

	.. versionchanged:: 2.0.0  ``encoder`` must now be a :class:`~.TomlEncoder` type or instance.
	.. latex:clearpage::
	"""

	if isinstance(encoder, type):
		encoder = encoder()

	return ''.join(encoder.dumps(data, name=''))


def dump(
		data: Mapping[str, Any],
		filename: PathLike,
		encoder: Union[Type[TomlEncoder], TomlEncoder] = TomlEncoder,
		) -> str:
	r"""
	Writes out ``data`` as TOML to the given file.

	:param data:
	:param filename: The filename to write to.
	:param encoder: The :class:`~.TomlEncoder` to use for constructing the output string.

	:returns: A string containing the ``TOML`` corresponding to ``data``.

	.. versionchanged:: 2.0.0  ``encoder`` must now be a :class:`~.TomlEncoder` type or instance.
	"""

	filename = PathPlus(filename)
	as_toml = dumps(data, encoder=encoder)
	filename.write_clean(as_toml)
	return as_toml


def loads(
		s: str,
		decoder: Union[Type[TomlDecoder], TomlDecoder] = TomlDecoder,
		) -> Dict[str, Any]:
	r"""
	Parse the given string as TOML.

	:param s:
	:param decoder: The :class:`~.TomlEncoder` to use for constructing the output string.

	:returns: A mapping containing the ``TOML`` data.

	.. versionchanged:: 2.0.0  ``decoder`` must now be a :class:`~.TomlDecoder` type or instance.
	"""

	if not isinstance(s, str):
		raise TypeError("Expecting something like a string")

	if isinstance(decoder, type):
		decoder = decoder()

	return decoder.loads(s)


def load(
		filename: PathLike,
		decoder: Union[Type[TomlDecoder], TomlDecoder] = TomlDecoder,
		) -> Dict[str, Any]:
	r"""
	Parse TOML from the given file.

	:param filename: The filename to read from to.
	:param decoder: The :class:`~.TomlEncoder` to use for constructing the output string.

	:returns: A mapping containing the ``TOML`` data.

	.. versionchanged:: 2.0.0  ``decoder`` must now be a :class:`~.TomlDecoder` type or instance.
	"""

	return loads(
			PathPlus(filename).read_text(),
			decoder=decoder,
			)



#!/usr/bin/env python3
#
#  encoder.py
"""
Dom's custom encoder for Tom's Obvious, Minimal Language.

.. versionadded:: 0.2.0
"""
#
#  Copyright © 2021 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Based on https://github.com/hukkin/tomli-w
#  MIT Licensed
#  Copyright (c) 2021 Taneli Hukkinen
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import pathlib
import string
from datetime import date, datetime, time
from decimal import Decimal
from types import MappingProxyType
from typing import Any, Dict, Iterator, List, Mapping, Tuple, Union

# this package
from dom_toml.decoder import InlineTableDict

__all__ = ["TomlEncoder", "TomlArraySeparatorEncoder", "TomlNumpyEncoder", "TomlPathlibEncoder"]

ASCII_CTRL = frozenset(chr(i) for i in range(32)) | frozenset(chr(127))
ILLEGAL_BASIC_STR_CHARS = frozenset('"\\') | ASCII_CTRL - frozenset('\t')
BARE_KEY_CHARS = frozenset(string.ascii_letters + string.digits + "-_")
ARRAY_TYPES = (list, tuple)
ARRAY_INDENT = ' ' * 4

COMPACT_ESCAPES = MappingProxyType({
		'\x08': "\\b",  # backspace
		'\n': "\\n",  # linefeed
		'\x0c': "\\f",  # form feed
		'\r': "\\r",  # carriage return
		'"': '\\"',  # quote
		'\\': "\\\\",  # backslash
		})


class TomlEncoder:
	"""
	TOML encoder which wraps long lists onto multiple lines and adds a blank line before arrays of tables.

	:param preserve:
	:param allow_multiline:
	:param separator:

	.. versionchanged:: 0.2.0  Moved from ``__init__.py``
	.. versionchanged:: 2.0.0  Added ``allow_multiline``  argument.
	.. autosummary-widths:: 45/100
	"""

	# The maximum width of the list **value**, after which it will be wrapped.
	max_width: int = 100

	allow_multiline: bool

	# cache rendered inline tables (mapping from object id to rendered inline table)
	inline_table_cache: Dict[int, str]

	def __init__(self, preserve: bool = True, multiline_strings: bool = False):
		self.preserve = preserve
		self.allow_multiline = multiline_strings
		self.inline_table_cache = {}

	def dumps(
			self,
			table: Mapping[str, Any],
			*,
			name: str,
			inside_aot: bool = False,
			) -> Iterator[str]:
		"""
		Serialise the given table.

		:param name: The table name.
		:param inside_aot:

		:rtype:

		.. versionadded:: 2.0.0
		"""

		yielded = False
		literals = []
		tables: List[Tuple[str, Any, bool]] = []  # => [(key, value, inside_aot)]
		for k, v in table.items():
			if v is None:
				continue
			elif self.preserve and isinstance(v, InlineTableDict):
				literals.append((k, v))
			elif isinstance(v, dict):
				tables.append((k, v, False))
			elif self._is_aot(v) and not all(self._is_suitable_inline_table(t) for t in v):
				tables.extend((k, t, True) for t in v)
			else:
				literals.append((k, v))

		if inside_aot or name and (literals or not tables):
			yielded = True
			yield f"[[{name}]]\n" if inside_aot else f"[{name}]\n"

		if literals:
			yielded = True
			for k, v in literals:
				yield f"{self.format_key_part(k)} = {self.format_literal(v)}\n"

		for k, v, in_aot in tables:
			if yielded:
				yield '\n'
			else:
				yielded = True

			key_part = self.format_key_part(k)
			display_name = f"{name}.{key_part}" if name else key_part

			yield from self.dumps(v, name=display_name, inside_aot=in_aot)

	def format_literal(self, obj: object, *, nest_level: int = 0) -> str:
		"""
		Format a literal value.

		:param obj:
		:param nest_level:

		:rtype:

		.. versionadded:: 2.0.0
		"""

		if isinstance(obj, bool):
			return "true" if obj else "false"
		if isinstance(obj, (int, float, date, datetime)):
			return str(obj)
		if isinstance(obj, Decimal):
			return self.format_decimal(obj)
		if isinstance(obj, time):
			if obj.tzinfo:
				raise ValueError("TOML does not support offset times")
			return str(obj)
		if isinstance(obj, str):
			return self.format_string(obj, allow_multiline=self.allow_multiline)
		if isinstance(obj, ARRAY_TYPES):
			return self.format_inline_array(obj, nest_level)
		if isinstance(obj, dict):
			return self.format_inline_table(obj)
		raise TypeError(f"Object of type {type(obj)} is not TOML serializable")

	def format_decimal(self, obj: Decimal) -> str:
		"""
		Format a decimal value.

		:param obj:

		:rtype:

		.. versionadded:: 2.0.0
		"""

		if obj.is_nan():
			return "nan"
		if obj == Decimal("inf"):
			return "inf"
		if obj == Decimal("-inf"):
			return "-inf"
		return str(obj)

	def format_inline_table(self, obj: dict) -> str:
		"""
		Format an inline table.

		:param obj:

		:rtype:

		.. versionadded:: 2.0.0
		"""

		# check cache first
		obj_id = id(obj)
		if obj_id in self.inline_table_cache:
			return self.inline_table_cache[obj_id]

		if not obj:
			rendered = "{}"
		else:
			rendered = (
					"{ "
					+ ", ".join(f"{self.format_key_part(k)} = {self.format_literal(v)}" for k, v in obj.items())
					+ " }"
					)
		self.inline_table_cache[obj_id] = rendered
		return rendered

	def format_inline_array(self, obj: Union[Tuple, List], nest_level: int) -> str:
		"""
		Format an inline array.

		:param obj:
		:param nest_level:

		:rtype:

		.. versionadded:: 2.0.0
		"""

		if not len(obj):
			return "[]"

		item_indent = ARRAY_INDENT * (1 + nest_level)
		closing_bracket_indent = ARRAY_INDENT * nest_level
		single_line = "[ " + ", ".join(
				self.format_literal(item, nest_level=nest_level + 1) for item in obj
				) + f",]"

		if len(single_line) <= self.max_width:
			return single_line
		else:
			start = "[\n"
			body = ",\n".join(item_indent + self.format_literal(item, nest_level=nest_level + 1) for item in obj)
			end = f",\n{closing_bracket_indent}]"
			return start + body + end

	def format_key_part(self, part: str) -> str:
		"""
		Format part of a key.

		:param part:

		:rtype:

		.. versionadded:: 2.0.0
		"""

		if part and BARE_KEY_CHARS.issuperset(part):
			return part
		return self.format_string(part, allow_multiline=False)

	def format_string(self, s: str, *, allow_multiline: bool) -> str:
		"""
		Format a string.

		:param s:
		:param allow_multiline:

		:rtype:

		.. versionadded:: 2.0.0
		.. latex:clearpage::
		"""

		do_multiline = allow_multiline and '\n' in s
		if do_multiline:
			result = '"""\n'
			s = s.replace("\r\n", '\n')
		else:
			result = '"'

		pos = seq_start = 0
		while True:
			try:
				char = s[pos]
			except IndexError:
				result += s[seq_start:pos]
				if do_multiline:
					return result + '"""'
				return result + '"'
			if char in ILLEGAL_BASIC_STR_CHARS:
				result += s[seq_start:pos]
				if char in COMPACT_ESCAPES:
					if do_multiline and char == '\n':
						result += '\n'
					else:
						result += COMPACT_ESCAPES[char]
				else:
					result += "\\u" + hex(ord(char))[2:].rjust(4, '0')
				seq_start = pos + 1
			pos += 1

	def _is_aot(self, obj: Any) -> bool:
		"""
		Decides if an object behaves as an array of tables (i.e. a nonempty list of dicts).

		:param obj:
		"""

		return bool(isinstance(obj, ARRAY_TYPES) and obj and all(isinstance(v, dict) for v in obj))

	def _is_suitable_inline_table(self, obj: dict) -> bool:
		"""
		Use heuristics to decide if the inline-style representation is a good choice for a given table.

		:param obj:
		"""

		# if self.preserve and isinstance(dict, InlineTableDict):
		# 	return True

		rendered_inline = f"{ARRAY_INDENT}{self.format_inline_table(obj)},"
		return len(rendered_inline) <= self.max_width and '\n' not in rendered_inline


class TomlPathlibEncoder(TomlEncoder):
	"""
	TOML Encoder with pathlib support.

	.. versionadded:: 2.0.0
	"""

	def format_literal(self, obj: object, *, nest_level: int = 0) -> str:
		"""
		Format a literal value.

		:param obj:
		:param nest_level:
		"""

		if isinstance(obj, pathlib.PurePath):
			obj = str(obj)
		return super().format_literal(obj, nest_level=nest_level)


class TomlNumpyEncoder(TomlEncoder):
	"""
	TOML Encoder with support for numpy types.

	.. versionadded:: 2.0.0
	"""

	def format_literal(self, obj: object, *, nest_level: int = 0) -> str:
		"""
		Format a literal value.

		:param obj:
		:param nest_level:
		"""

		# 3rd party
		import numpy as np  # nodep

		if isinstance(obj, (np.float16, np.float32, np.float64)):
			return self._dump_float(obj)  # type: ignore[arg-type]

		elif isinstance(obj, (np.int16, np.int32, np.int64)):
			return self._dump_int(obj)  # type: ignore[arg-type]

		elif isinstance(obj, np.ndarray):
			return self.format_inline_array(obj, nest_level)  # type: ignore[arg-type]

		return super().format_literal(obj, nest_level=nest_level)

	def _dump_int(self, v: int) -> str:
		return f"{int(v)}"

	def _dump_float(self, v: float) -> str:
		return f"{v}".replace("e+0", "e+").replace("e-0", "e-")


class TomlArraySeparatorEncoder(TomlEncoder):
	"""
	TOML Encoder with adjustable array separator.

	:param preserve:
	:param allow_multiline:
	:param separator:

	.. versionadded:: 2.0.0
	"""

	def __init__(self, preserve: bool = True, multiline_strings: bool = False, separator: str = ','):
		self.preserve = preserve
		self.allow_multiline = multiline_strings
		self.inline_table_cache = {}

		if not separator.strip():
			separator = ',' + separator
		elif separator.strip(' \t\n\r,'):
			raise ValueError("Invalid separator for arrays")

		self.separator = separator

	def format_inline_array(self, obj: Union[Tuple, List], nest_level: int) -> str:
		"""
		Format an inline array.

		:param obj:
		:param nest_level:
		"""

		t = []
		retval = '['
		for u in obj:
			t.append(self.format_literal(u, nest_level=nest_level))
		while t != []:
			s: List[str] = []
			for u in t:
				if isinstance(u, list):
					s.extend(r for r in u)
				else:
					retval += ' ' + str(u) + self.separator
			t = s
		retval += ']'
		return retval
